#define DBUS_API_SUBJECT_TO_CHANGE
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#define DbusDomain "org.freedesktop.compiz"



static DBusConnection * DbusConnection = NULL;

gboolean init_dbus (void)
{
	DBusError error;
	
	if (DbusConnection != NULL)
		return TRUE;
	
	dbus_error_init (&error);
	if (!(DbusConnection = dbus_bus_get (DBUS_BUS_SESSION, &error))) {
		g_warning ("could not get system bus: %s\n", error.message);
		dbus_error_free (&error);
		return FALSE;
	}
	
	dbus_connection_setup_with_g_main (DbusConnection, NULL);
	dbus_connection_set_exit_on_disconnect (DbusConnection, FALSE);
	
	DBusMessage     *message, * reply;
	dbus_error_init (&error);
	message = dbus_message_new_method_call (DbusDomain,
 						"/org/freedesktop/compiz/core/allscreens/audible_bell",
						DbusDomain,
						"get");

	reply = dbus_connection_send_with_reply_and_block (DbusConnection, 
							   message,
							   -1,
							   &error);
	
	if (!reply) {
		//printf("Error: %s Path: %s\n", error.message);
		//g_print ("Error while getting drives from daemon.\n");
		dbus_message_unref (message);
		return FALSE;
	}

     return TRUE;
}

  GSList * get_string_list(DBusMessageIter *iter)
      {
      	GSList * list = NULL;
      
      	while (dbus_message_iter_get_arg_type(iter) == DBUS_TYPE_STRING) {
      		const char *value;
      		char *t;
      		dbus_message_iter_get_basic(iter, &value);
      		t = g_strdup(value);
		list = g_slist_append(list, t);
      		dbus_message_iter_next(iter);
      	}

      	return list;
  }


  char ** get_string_array(DBusMessageIter *iter, char *extra)
      {
      	GArray *array;
      	char **result;
      	array = g_array_new(TRUE, FALSE, sizeof(char *));
      
      	while (dbus_message_iter_get_arg_type(iter) == DBUS_TYPE_STRING) {
      		const char *value;
      		char *t;
      		dbus_message_iter_get_basic(iter, &value);
      		t = g_strdup(value);
      		g_array_append_vals(array, &t, 1);
      		dbus_message_iter_next(iter);
      	}
      	if (extra != NULL)
      		g_array_append_vals(array, &extra, 1);
      	result = (char **) array->data;
      	g_array_free(array, FALSE);
      	return result;
      }


/* Returns a list of CompizSettingsOption's for any method, upto the calling function to interpret it */
GSList * get_dbus_options_list (gchar * plugin, gchar * screen, gchar * key, gchar * method)
{


	DBusMessage     *message, *reply;
	GSList           * list;

	gchar		* dbus_path;
	DBusMessageIter  iter;
        DBusError error;
	int current_type;

	if (!strcmp(plugin, "CompizSettings_general")) {
   		dbus_path = g_strdup_printf("/org/freedesktop/compiz/core/%s/%s", screen, key);
	} else {
   		dbus_path = g_strdup_printf("/org/freedesktop/compiz/%s/%s/%s", plugin, screen, key);
	}

	dbus_error_init (&error);
	message = dbus_message_new_method_call (DbusDomain,
 						dbus_path,
						DbusDomain,
						method);

	reply = dbus_connection_send_with_reply_and_block (DbusConnection, 
							   message,
							   -1,
							   &error);

	
	if (!reply) {
		//printf("Error: %s Path: %s\n", error.message, dbus_path);
		//g_print ("Error while getting drives from daemon.\n");
		dbus_message_unref (message);
		return NULL;
	}

	list = NULL;	

	dbus_message_iter_init (reply, &iter);

	double float_value;
	gpointer * data;

 	while ((current_type = dbus_message_iter_get_arg_type (&iter)) != DBUS_TYPE_INVALID) {
	CompizSettingsOption * option = malloc(sizeof(CompizSettingsOption));

      if (dbus_type_is_basic(dbus_message_iter_get_arg_type (&iter)) == TRUE) {
	    	  dbus_message_iter_get_basic (&iter, &data);
	    	  option->data = data;

	  	if(dbus_message_iter_get_arg_type (&iter) == DBUS_TYPE_INT32) {
	    	  option->type = CompizSettingsOptionInt;
	  	} else if (dbus_message_iter_get_arg_type (&iter) == DBUS_TYPE_DOUBLE) {
	   	  /* Can't store float as pointer */
	 	  dbus_message_iter_get_basic (&iter, &float_value);
	          option->data = (gpointer *) g_strdup_printf("%f", float_value);
	    	  option->type = CompizSettingsOptionFloat;
	  	} else if (dbus_message_iter_get_arg_type (&iter) == DBUS_TYPE_STRING) {
	    	  option->type = CompizSettingsOptionString;
	  	}

	} else if (dbus_message_iter_get_arg_type (&iter) == DBUS_TYPE_ARRAY) {
	     DBusMessageIter sub_iter;
	      dbus_message_iter_recurse(&iter, &sub_iter);
	      option->data = (gpointer *) get_string_list(&sub_iter);

	    option->type = CompizSettingsOptionStringList;
        }


	 list = g_slist_append(list, option);
   	  dbus_message_iter_next (&iter);
	}



   return list;

   
}

CompOptionIntRestriction * backend_get_int_restriction(gchar * plugin, gchar * screen, gchar * key) {

   CompOptionIntRestriction * int_restriction = malloc(sizeof(CompOptionIntRestriction));
   gboolean restriction_got = FALSE;
   int number = 0;

  GSList * options_list = get_dbus_options_list(plugin, screen, key, "getMetadata");
		for (; options_list; options_list = options_list->next) {
		   CompizSettingsOption * tmp_option = options_list->data;
		    if (tmp_option->type == CompizSettingsOptionInt) {
			if (number == 3) {
			  int_restriction->min = (int) tmp_option->data;
			} else if (number == 4) {
			  int_restriction->max = (int) tmp_option->data;
			  restriction_got = TRUE;
			}
		    }
		  number++;
		}

    if (restriction_got == TRUE) {
       return int_restriction;
    } else {
       return NULL;
    }
}

CompOptionFloatRestriction * backend_get_float_restriction(gchar * plugin, gchar * screen, gchar * key) {

   CompOptionFloatRestriction * float_restriction = malloc(sizeof(CompOptionFloatRestriction));
   
   gboolean restriction_got = FALSE;
   int number = 0;

  GSList * options_list = get_dbus_options_list(plugin, screen, key, "getMetadata");
		for (; options_list; options_list = options_list->next) {

		   CompizSettingsOption * tmp_option = options_list->data;
		    if (tmp_option->type == CompizSettingsOptionFloat) {
			if (number == 3) {
	   		    float float_value = (float) g_strtod ((gchar *) tmp_option->data, NULL);
			    float_restriction->min = float_value;
			} else if (number == 4) {
	   		    float float_value = (float) g_strtod ((gchar *) tmp_option->data, NULL);
			    float_restriction->max = float_value;
			} else if (number == 5) {
	   		    float float_value = (float) g_strtod ((gchar *) tmp_option->data, NULL);
			    float_restriction->precision = float_value;
			    restriction_got = TRUE;
			}
		    }
		  number++;
		}

    if (restriction_got == TRUE) {
       return float_restriction;
    } else {
       return NULL;
    }

}


GSList * backend_get_string_restriction(gchar * plugin, gchar * screen, gchar * key) {

   GSList * string_restriction = NULL;
   
   gboolean restriction_got = FALSE;
   int number = 0;

  GSList * options_list = get_dbus_options_list(plugin, screen, key, "getMetadata");
		for (; options_list; options_list = options_list->next) {
		   CompizSettingsOption * tmp_option = options_list->data;

			if (number > 2) {
			    if (tmp_option->type == CompizSettingsOptionStringList) {
				restriction_got = TRUE;
				return (GSList *)tmp_option->data;
			    } else if (tmp_option->type == CompizSettingsOptionString) {
				string_restriction = g_slist_append(string_restriction, tmp_option->data);
				restriction_got = TRUE;
			    }
			}
		  number++;
		}

    if (restriction_got == TRUE) {
       return string_restriction;
    } else {
       return NULL;
    }

}


gint dbus_backend_get_option_type(gchar * plugin, gchar * screen, gchar * key) {

   int number = 0;
   int OptionType = 0;

  GSList * options_list = get_dbus_options_list(plugin, screen, key, "getMetadata");
		for (; options_list; options_list = options_list->next) {
		   CompizSettingsOption * tmp_option = options_list->data;

			if (number == 2) {
			  if (tmp_option->type == CompizSettingsOptionString) {
			    if (!strcmp((gchar *) tmp_option->data,  "color")) {
				OptionType = CompizSettingsOptionColor;
			    } else if  (!strcmp((gchar *) tmp_option->data,  "bool")) {
				OptionType = CompizSettingsOptionBool;
			    } else if  (!strcmp((gchar *) tmp_option->data,  "string")) {
				OptionType = CompizSettingsOptionString;
			    } else if  (!strcmp((gchar *) tmp_option->data,  "int")) {
				OptionType = CompizSettingsOptionInt;
			    } else if  (!strcmp((gchar *) tmp_option->data,  "float")) {
				OptionType = CompizSettingsOptionFloat;
			    } else if  (!strcmp((gchar *) tmp_option->data,  "action")) {
				OptionType = CompizSettingsOptionAction;
			    }
			  }
			}
		  number++;
		}


    return OptionType;

}


