/* All back end specific stuff should go here, so changes in future can easily be made e.g. switching to dbus */

#include <gconf/gconf-client.h>
#define COMPIZ_PATH "/apps/compiz"
#define COMPIZ_PLUGINS_PATH COMPIZ_PATH "/plugins"
#define PROGRAM_PATH "/apps/compiz-settings"
#define PROGRAM_PLUGINS_PATH PROGRAM_PATH "/plugins"

 GConfClient *client;

enum {
  CompizSettingsOptionString = 1,
  CompizSettingsOptionStringList,
  CompizSettingsOptionInt,
  CompizSettingsOptionFloat,
  CompizSettingsOptionBool,
  CompizSettingsOptionColor,
  CompizSettingsOptionAction,
  CompizSettingsOptionIntORFloatORBoolORString /* mother of all hacks, should only be temporary though*/
};


enum {
  EdgeTopLeft = 1,
  EdgeTop,
  EdgeTopRight,
  EdgeRight,
  EdgeBottomRight,
  EdgeBottom,
  EdgeBottomLeft,
  EdgeLeft,
};

gchar * action_extentions[]={
    "_bell",
    "_key",
    "_button",
    "_edgebutton",
    "_edge",
};

//Bit of a gconf hack, got to keep a list of all gconf callbacks which have been setup because when you detroy a gtk widget and then create new ones it doesn't play nice unless the callbacks are first removed. This can be removed when moved to dbus i think
GSList * gconf_notify_ints;

void remove_gconf_notifier(gint *data) {
   gconf_client_notify_remove (client, (int) data);
   gconf_notify_ints =g_slist_remove (gconf_notify_ints, data);
}

void backened_remove_callbacks () {
 g_slist_foreach(gconf_notify_ints, (GFunc) remove_gconf_notifier, NULL);

}

/* End */

gboolean backend_is_setup(void) {
		  GConfEntry *entry;
		  entry = gconf_client_get_entry (client, COMPIZ_PATH "/general/allscreens/options/active_plugins", NULL, TRUE, NULL);

		   if (entry->value != NULL) {
			gconf_entry_free(entry);
			return TRUE;
		   } else {
			gconf_entry_free(entry);
			return FALSE;
		   }
}



void backend_initial_setup() {
  client = gconf_client_get_default ();
    gconf_client_add_dir (client,
			  COMPIZ_PATH,
			  GCONF_CLIENT_PRELOAD_RECURSIVE,
			  NULL);

    gconf_client_add_dir (client,
			  PROGRAM_PATH,
			  GCONF_CLIENT_PRELOAD_RECURSIVE,
			  NULL);
}


void backend_set_option(gchar * plugin, gchar * screen, gchar * key, gpointer * data, int * type) {

     gchar * path;

     if (!strcmp("CompizSettings_general", plugin)) {
          path = g_strdup_printf(COMPIZ_PATH "/general/%s/options/%s", screen, key);
	} else {
          path = g_strdup_printf(COMPIZ_PLUGINS_PATH "/%s/%s/options/%s", plugin, screen, key);
	}


	if (type == GINT_TO_POINTER(CompizSettingsOptionString)) {
	 gconf_client_set_string (client, path, (gchar *) data, NULL);
	} else if (type == GINT_TO_POINTER(CompizSettingsOptionColor)) {
	 gconf_client_set_string (client, path, (gchar *) data, NULL);
	} else if (type == GINT_TO_POINTER(CompizSettingsOptionAction)) {
	  CompizSettingsActionKey * option = (CompizSettingsActionKey *) data;
	    /* Key Shortcut */
	    gchar * key_path = g_strdup_printf("%s_key", path);
	    gconf_client_set_string (client, key_path, (gchar *) option->key, NULL);
	    g_free(key_path);

	    /* Mouse Shortcut */
	    gchar * button_path = g_strdup_printf("%s_button", path);
	    gconf_client_set_string (client, button_path, (gchar *) option->button, NULL);
	    g_free(button_path);

	    /* Screen Edges */
	    gchar * edge_path = g_strdup_printf("%s_edge", path);
	    gconf_client_set_list (client, edge_path, GCONF_VALUE_STRING, (GSList *) option->edge, NULL);
	    g_free(edge_path);

	} else if (type == GINT_TO_POINTER(CompizSettingsOptionBool)) {
	 gconf_client_set_bool (client, path, (gboolean) data, NULL);
	} else if (type == GINT_TO_POINTER(CompizSettingsOptionInt)) {
	 gconf_client_set_int (client, path, (gint) data, NULL);
	} else if (type == GINT_TO_POINTER(CompizSettingsOptionFloat)) {
	   float float_value = (float) g_strtod ((gchar *) data, NULL);
	   printf("Setting float %s, %f\n", path, float_value);
	   gconf_client_set_float (client, path, (float) float_value, NULL);
	} else if (type == GINT_TO_POINTER(CompizSettingsOptionStringList)) {
	 gconf_client_set_list (client, path, GCONF_VALUE_STRING, (GSList *) data, NULL);
	} else {
	 printf("Data type for %s is not supported", path);
	}	

    g_free(path);

}

gpointer * backend_get_option(gchar * plugin, gchar * screen, gchar * key, int type) {

     gchar * path;

     if (!strcmp("CompizSettings_general", plugin)) {
          path = g_strdup_printf(COMPIZ_PATH "/general/%s/options/%s", screen, key);
	} else {
          path = g_strdup_printf(COMPIZ_PLUGINS_PATH "/%s/%s/options/%s", plugin, screen, key);
	}

    gpointer * data = NULL;
	if (type == CompizSettingsOptionString) {
	 data = (gpointer *) gconf_client_get_string (client, path, NULL);
	} else if (type == CompizSettingsOptionColor) {
	 data = (gpointer *) gconf_client_get_string (client, path, NULL);
	} else if (type == CompizSettingsOptionAction) {
		CompizSettingsActionKey * action = malloc(sizeof(CompizSettingsActionKey));
		gchar * key_path = g_strdup_printf("%s_key", path);
		gchar * button_path = g_strdup_printf("%s_button", path);
		gchar * edge_path = g_strdup_printf("%s_edge", path);
		gchar * edge_button = g_strdup_printf("%s_edgebutton", path);

		  action->key =  (gchar *) gconf_client_get_string (client, key_path, NULL);
		  action->button =  (gchar *) gconf_client_get_string (client, button_path, NULL);
		  action->edge =  (GSList *) gconf_client_get_list (client, edge_path, GCONF_VALUE_STRING, NULL);
		  action->edge_button = (int) gconf_client_get_int (client, path, NULL);

		  g_free(key_path);
		  g_free(edge_path);
		  g_free(button_path);
		  g_free(edge_button);

	    data = (gpointer *) action;

	} else if (type == CompizSettingsOptionBool) {
	 data = (gpointer *) gconf_client_get_bool (client, path, NULL);
	} else if (type == CompizSettingsOptionInt) {
	 data = (gpointer *) gconf_client_get_int (client, path, NULL);
	} else if (type == CompizSettingsOptionFloat) {
	 data = (gpointer *) (int)(gconf_client_get_float (client, path, NULL) * 1000);
	} else if (type == CompizSettingsOptionStringList) {
	 data = (gpointer *) gconf_client_get_list (client, path, GCONF_VALUE_STRING, NULL);
	} 

     g_free(path);

    return data;
}


gint backend_is_action_key(gchar * full_path) {
/* I have speperated this becaause with gconf there is no easily way to detect an action key and i wanted to keep backend_get_option_type() as clean as possible, with dbus it can probably be removed */
gchar * key = g_strdup(full_path);


int i = 0;
while (i <= 4) {
	key = (gchar *) str_replace(key, action_extentions[i], "");
i++;
}



gboolean all_other_extensions_exist(gchar *key) {
  i = 0;
	while (i <= 4) {
	   GConfValue * entry;
	   gchar * tmp = g_strdup_printf("%s%s", key, action_extentions[i]);
	   entry = gconf_client_get (client, tmp, NULL);
	   //i != 3 is a backwards compatability hack for schema's that don't have edgebutton
	   if(entry == NULL && i != 3) {
	     return FALSE;
	   }
	   g_free(tmp);
	  	
		if (entry != NULL) {
	   		gconf_value_free(entry);
		}

		i++;
	}
   return TRUE;
}

  /* Almost certainly an action key, but not reliable enough */
  	if (all_other_extensions_exist(key) == TRUE) {
  	/*Certainly an action key */
	  return CompizSettingsOptionAction;
  	}

  return (int)NULL;

}

GSList * backend_get_list_of_plugins() {
  GSList * Plugins_list;
  GSList * new_list = NULL;
  /* Whatever backended used this function should return a lower case gslist of all available compiz plugins */
   Plugins_list = gconf_client_all_dirs (client, COMPIZ_PLUGINS_PATH, NULL);

  while (Plugins_list)
    {
      new_list = g_slist_append(new_list, g_path_get_basename(Plugins_list->data));
      Plugins_list = Plugins_list->next;
    }

   g_slist_free(Plugins_list);
   return new_list;
}

GSList * backend_get_list_of_active_plugins() {
  GSList * Plugins_list;
  GSList * new_list = NULL;
  /* Whatever backended used this function should return a lower case gslist of all available compiz plugins */
   Plugins_list = (GSList *) backend_get_option("CompizSettings_general", "allscreens", "active_plugins", CompizSettingsOptionStringList);

  while (Plugins_list)
    {
      new_list = g_slist_append(new_list, g_path_get_basename(Plugins_list->data));
      Plugins_list = Plugins_list->next;
    }

   g_slist_free(Plugins_list);
   return new_list;
}

gint * backend_get_option_type(gchar * plugin, gchar * screen, gchar * key) {

 gchar * path;


     if (!strcmp("CompizSettings_general", plugin)) {
          path = g_strdup_printf(COMPIZ_PATH "/general/%s/options/%s", screen, key);
	} else {
          path = g_strdup_printf(COMPIZ_PLUGINS_PATH "/%s/%s/options/%s", plugin, screen, key);
	}

    GConfValue * value;
    int dbus_value = dbus_backend_get_option_type(plugin, screen, key);
    int option = 0;


	value = gconf_client_get (client, path, NULL);
	gint action_key_tmp = (gint) backend_is_action_key(path);

	if (dbus_value == CompizSettingsOptionColor) {
	 option = CompizSettingsOptionColor;
	} else if (dbus_value == CompizSettingsOptionBool) {
	 option = CompizSettingsOptionBool;
	} else if (dbus_value == CompizSettingsOptionString) {
	 option = CompizSettingsOptionString;
	} else if (dbus_value == CompizSettingsOptionInt) {
	 option = CompizSettingsOptionInt;
	} else if (dbus_value == CompizSettingsOptionFloat) {
	 option = CompizSettingsOptionFloat;
	} else if (dbus_value == CompizSettingsOptionAction) {
	 option = CompizSettingsOptionAction; 
	}else if (action_key_tmp == CompizSettingsOptionAction) {
	 option = CompizSettingsOptionAction; 
	} else if (value->type == GCONF_VALUE_STRING) {
	 option = CompizSettingsOptionString; 
	} else if (value->type == GCONF_VALUE_INT) {
	 option = CompizSettingsOptionInt; 
	} else if (value->type == GCONF_VALUE_FLOAT) {
	 option = CompizSettingsOptionFloat; 
	} else if (value->type == GCONF_VALUE_BOOL) {
	 option = CompizSettingsOptionBool; 
	} else {
	  // printf("     !! Value type %s unknown compiz-settings is still in development\n", path);
	}


	g_free(path);

	if (value != NULL) {
    	  gconf_value_free(value);
	}


  return (gint *)option;
}

GSList * backend_get_list_of_options(gchar * plugin_name, gchar * screen) {

//printf("backend_get_list_of_options Getting list of options \n");

  GSList * list = NULL;
  GSList * tmplist = NULL;
  GSList * new_list = NULL;
  gchar * key;


   if (!strcmp(plugin_name, "CompizSettings_general")) {
    key = g_strdup_printf(COMPIZ_PATH "/general/%s/options", screen);
   } else {
    key = g_strdup_printf(COMPIZ_PLUGINS_PATH "/%s/%s/options", plugin_name, screen);
   }

   list = gconf_client_all_entries (client, key, NULL);

   tmplist = list;
   while(list) {
	GConfEntry * entry = list->data;
	   CompizSettingsOption * option = malloc(sizeof(CompizSettingsOption));
	    option->key = g_path_get_basename(entry->key);
	    option->plugin_name = g_strdup(plugin_name);
	    option->screen = g_strdup(screen);
	    option->type = (int) backend_get_option_type(plugin_name, screen, option->key);
	    option->data = NULL;

	  /* Hack for action keys because gconf reports 4 keys _edge _bell etc when only one is wanted */
	  if (option->type == CompizSettingsOptionAction) {
		gchar * key = option->key;
		int i = 0;
		 while (i <= 4) {
		  key = (gchar *) str_replace(key, action_extentions[i], "");
		  i++;
		 }
		  if(!key_is_in_option_list(new_list, key)) {
		      option->key = key;
	 	      //list->data = option;
	    		option->data = (gpointer *)NULL;
	    		option->data = backend_get_option(plugin_name, screen, option->key, option->type);
	 	    new_list = g_slist_append((GSList *)new_list, (gpointer *) option);
		  }
	   } else {
	    	        option->data = (gpointer *)NULL;
	    		option->data = backend_get_option(plugin_name, screen, option->key, option->type);
	 	new_list = g_slist_append((GSList *)new_list,  (gpointer *)option);
	   }

		list = list->next;

	 gconf_entry_free(entry);

    }
   list = tmplist;

   g_slist_free(list);
   g_free(key);

  return new_list;
}


//The next few functions are going to be the hardest to convert 

void backend_create_new_callback(gchar * key, GFunc * callback_function, GtkWidget * widget) {
	/* Create a new callback to update changes when a widget value is changed */
	gint * notify_int;
	notify_int = (gint *) gconf_client_notify_add (client,
				 key,
				 (GConfClientNotifyFunc)callback_function,
				 widget,
				 NULL,
				 NULL);


	//Hack see gconf_notify_ints
	gconf_notify_ints = g_slist_append (gconf_notify_ints, GINT_TO_POINTER (notify_int)); 
	
	//g_free(notify_int);
}

  /* Create a new callback to notify the widget when settings changes are made external. Not sure if this functionality exists in dbus so mostly gconf specific */
void backend_setting_changed_externally(GConfClient *client,
		    guint        signal,
		    GConfEntry  *entry,
		    gpointer * user_data) {

	GConfValue *value;
	value = gconf_entry_get_value (entry); 
	GtkWidget  *widget = GTK_WIDGET (user_data);
	gchar * widget_type = (gchar *) GTK_OBJECT_TYPE_NAME(widget);

	if (!strcmp(widget_type, "GtkEntry")) {
	  	gchar * new_string;
		 new_string = (gchar *) gconf_value_get_string (value);
		 gtk_entry_set_text(GTK_ENTRY (widget), new_string);
	} else  if (!strcmp(widget_type, "GtkCheckButton")) {
		gboolean bool;
		bool = gconf_value_get_bool (value);
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), bool);
	} else if (!strcmp(widget_type, "GtkHScale")) {
		if (gtk_scale_get_digits(GTK_SCALE(widget)) == 0) {
    	    	  int numb = gconf_value_get_int (value);
	    	  gtk_range_set_value (GTK_RANGE(widget), numb);
	 	} else if (gtk_scale_get_digits(GTK_SCALE(widget)) == 0) {
    	    	  float numb = gconf_value_get_float (value);
	    	  gtk_range_set_value (GTK_RANGE(widget), numb);
	 	}
  	 } else if (!strcmp(widget_type, "GtkComboBox")) {
		/* Needs work GtkTreeModel * model;
		GtkTreeIter * iter = malloc(sizeof(GtkTreeIter));
		 model = gtk_combo_box_get_model (widget);

		if (gtk_tree_model_get_iter_first   (model, iter)) {
		  while(gtk_tree_model_iter_next (model, iter)) {
			gchar * string = gtk_tree_model_get_string_from_iter (model, iter);
			printf("String %s\n", string);
		  }
		} */

	} else {
	 printf("Widget %s not supported\n", widget_type);
	}


	//For some reason this causes a seg fault when someone types too fast 
	//gconf_value_free(value); 


}

